# frozen_string_literal: true

# Released under the MIT License.
# Copyright, 2022-2024, by Samuel Williams.

source "https://rubygems.org"

gem "async"
gem "async-redis"
gem "async-pool", path: "../../"
