# frozen_string_literal: true

# Released under the MIT License.
# Copyright, 2019-2024, by Samuel Williams.

require_relative "pool/version"
require_relative "pool/controller"

# @namespace
module Async
	# @namespace
	module Pool
	end
end
