# frozen_string_literal: true

# Released under the MIT License.
# Copyright, 2021-2024, by Samuel Williams.

source "https://rubygems.org"

# Takes about ~10s
# gem "async-pool", "0.3.3"

# Takes about ~30s
gem "async-pool", path: "../../"
